// Dual-ported maze tracker RAM module
// one set of read-only access ports, one set of read/write.
// by Sam Revitch, CSE 467

// Theory and Purpose:
// A maze in general can be represented as a connected graph, where each
// vertex is a "room", which connects to other rooms via edges.
// Represented as squares in 2D euclidean space, each room has the possibility
// of connecting to one of its four neighbors.  Likewise, since the maze
// is undirected, connections are shared, and connected neighbors have
// return connections.

// The point of the maze keeper is to keep track of an NxM grid of rooms,
// and the accompanying shared connectivity list.

// To make this maximally useful to devices responsible for using the data,
// the maze keeper unit will accept row and column inputs for "room"
// addressing, and will return four connectivity bits, representing whether
// there is a path between the room and:
// bits[0] = top neighbor
// bits[1] = right neighbor
// bits[2] = bottom neighbor
// bits[3] = left neighbor

// Obviously, each room shares a few bits with its immediate neighbors. :-)
// e.g. if room 2,2 is written { 1,0,0,1 }, then:
// /-----------------------\
// |       |       |       |
// |  1,1  ?  1,2  ?  1,3  |
// |       |       |       |
// |---?------[1]------?---|
// |       |       |       |
// |  2,1 [1] 2,2 [0] 2,3  |
// |       |       |       |
// |---?------[0]------?---|
// |       |       |       |
// |  3,1  ?  3,2  ?  3,3  |
// |       |       |       |
// \-----------------------/

// Regarding connectivity of edge vertices (that is, vertices on the edges
// of the grid) mazekeeper always returns 0, and writing to these locations
// does nothing.

// All mazekeeper really does is to maintain an internal edge map and
// return pieces of it in terms of connected vertices.

// Mazekeeper outputs a valid result for each set of inputs at each
// positive clock edge, would have a relatively long setup time, and
// would have a relatively short propagation time.
// Mazekeeper starts out completely uninitialized and must be manually
// loaded with initializing data.

// Parameters:
// You can set the size of the grid managed by Mazekeeper, e.g.:
//
// mazekeeper #(32, 64, 5, 6) mymaze(blah, blah, ...);
//
// will make a 32-wide, 64-high maze grid.
// MAKE SURE to keep the WBITS and HBITS in sync with the width
// and height of the maze.

module mazekeeper(clk, row1, col1, bits1, row2, col2, bits2,
		  webits, din);

   // These parameters needn't be powers of 2.
   parameter WIDTH = 16, HEIGHT = 16;
   parameter WBITS = 4, HBITS = 4;

   input     clk;

   input [HBITS-1:0] row1, row2;
   input [WBITS-1:0] col1, col2;
   output [3:0]      bits1, bits2;

   input [3:0] 	     webits;         // per-bit write enable lines
   input [3:0] 	     din;

   // bits output registers
   reg [3:0] 	     bits1, bits2;

   wire [1:0] 	     row_bits1, col_bits1, row_bits2, col_bits2;
   wire [1:0] 	     row_we, col_we, row_data, col_data;

   // Note that there is some translation going on in the bus lines
   // The dp_memshifter units take the "top" bit as the zeroth, and the
   // "bottom" bit as the first, whereas the order is different in
   // din and webits.
   assign row_we = { webits[2], webits[0] };
   assign row_data = { din[2], din[0] };
   assign col_we = { webits[1], webits[3] };
   assign col_data = { din[1], din[3] };

   maze_dp_memshifter #(HEIGHT-1, WIDTH, HBITS, WBITS)
      rowmem(clk, row1, col1, row_bits1,
	     row2, col2, row_bits2,
	     row_we, row_data);
   
   // The column-to-column connectivity information is arranged identically
   // to the row-to-row info, except of course swapping the parameters. :-)
   maze_dp_memshifter #(WIDTH-1, HEIGHT, WBITS, HBITS)
      colmem(clk, col1, row1, col_bits1,
	     col2, row2, col_bits2,
	     col_we, col_data);

   // Copy results into bits1, bits2 at each positive clock edge
   always @(posedge clk) begin
      bits1 <= { col_bits1[0], row_bits1[1], col_bits1[1], row_bits1[0] };
      bits2 <= { col_bits2[0], row_bits2[1], col_bits2[1], row_bits2[0] };
   end // always @ (posedge clk)

endmodule // mazekeeper


// Dual-ported memory components for mazekeeper
// Separated in case xilinx gets other ideas about them,
// also I suppose it's possible to optimize.

// Memshifter operates two of the dp memory components outlined below,
// and reads/writes the bit at the specified position, plus the one in the
// next line but same column using odd/even addressing hacks.

module maze_dp_memshifter(clk, rd_line, rd_bit, rd_out,
  			       rw_line, rw_bit, rw_out, rw_we, rw_data);
   parameter LINES = 16, LWIDTH = 16, LBITS = 4, LWBITS = 4;

   input     clk;
   input [LBITS-1:0] rd_line, rw_line;
   input [LWBITS-1:0] rd_bit, rw_bit;
   output [1:0]       rd_out, rw_out;
   input [1:0] 	      rw_we, rw_data;
   
   wire [LBITS:0]     rd_linem1, rw_linem1;

   wire 	     out1o, out1e, out2o, out2e;
   wire 	     bit2ow, bit2od, bit2ew, bit2ed;

   assign rd_linem1 = rd_line - 1;
   assign rw_linem1 = rw_line - 1;
   
   assign bit2od = (rw_line[0] ? rw_data[0] : rw_data[1]);
   assign bit2ed = (rw_line[0] ? rw_data[1] : rw_data[0]);

   assign bit2ow = (rw_line[0] ? rw_we[0] : rw_we[1]);
   assign bit2ew = (rw_line == 0) ? 0 : (rw_line[0] ? rw_we[1] : rw_we[0]);

   always @(posedge clk) begin
//      $display("ed=%b od=%b ew=%b ow=%b, er=%d or=%d", bit2ed, bit2od,
//	       bit2ew, bit2ow, rw_linem1[LBITS-1:1], rw_line[LBITS-1:1]);
//      $display("er=%b or=%b bit=%d eo=%b oo=%b",
//	       rd_linem1[LBITS-1:0], rd_line[LBITS-1:0],
//	       rd_bit[LWBITS-1:0], out1e, out1o);
   end

   // Get the number of memory lines right
   // in the case of an odd total number of lines, the extra line belongs
   // to the odd memory cell.
   parameter ODDLINES = ((LINES % 2) ? ((LINES / 2) + 1) : (LINES / 2));
   parameter EVENLINES = (LINES / 2);

   // Note that the even memory unit has one wasted line.   
   maze_dp_mem #(EVENLINES, LWIDTH, LBITS-1, LWBITS)
      even_lines(clk,
		rd_linem1[LBITS-1:1], rd_bit[LWBITS-1:0], out1e,
		rw_linem1[LBITS-1:1], rw_bit[LWBITS-1:0], out2e,
		bit2ew, bit2ed);
   
   maze_dp_mem #(ODDLINES, LWIDTH, LBITS-1, LWBITS)
      odd_lines(clk,
	       rd_line[LBITS-1:1], rd_bit[LWBITS-1:0], out1o,
	       rw_line[LBITS-1:1], rw_bit[LWBITS-1:0], out2o,
	       bit2ow, bit2od);

   // Compute top(0) and bottom(1) bits
   assign rd_out[0] = (rd_line == 0) ? 0 :
		      (rd_line[0] ? out1o : out1e);
   assign rd_out[1] = (rd_line >= LINES) ? 0 :
		      (rd_line[0] ? out1e : out1o);

   assign rw_out[0] = (rw_line == 0) ? 0 :
		      (rw_line[0] ? out2o : out2e);
   assign rw_out[1] = (rw_line >= LINES) ? 0 :
		      (rw_line[0] ? out2e : out2o);

endmodule // maze_dp_shifter


// The actual dual ported row/column addressable bit memory unit
// optimize me if you dare

module maze_dp_mem(clk, rd_line, rd_bit, rd_out,
		        rw_line, rw_bit, rw_out, rw_we, rw_data);

   // These get overrided inside mazekeeper for row and column capacities
   parameter LINES = 16, LWIDTH = 16, LBITS = 4, LWBITS = 4;
   
   input clk;
   input [LBITS-1:0] rd_line, rw_line;
   input [LWBITS-1:0] rd_bit, rw_bit;
   output 	      rd_out, rw_out;
   input 	      rw_we, rw_data;

   wire 	      rd_out, rw_out;
   wire [LWIDTH-1:0]   rd_full_line, rw_full_line, rw_replace;

   reg [LWIDTH-1:0]   memory [LINES-1:0];

   reg [LINES-1:0]    testreg;   

   // Read-only data ports
   // Short circuit when a write is occurring
   assign rd_full_line = ((rw_line == rd_line) && rw_we) ?
			 rw_replace : memory[rd_line];
   assign rd_out = rd_full_line[rd_bit];

   // Read-write data port reading
   assign rw_full_line = memory[rw_line];
   assign rw_out = rw_full_line[rw_bit];

   // Read-write data port writing
   // First generate a replacement bitvector
   assign rw_replace = (rw_full_line & ~(1 << rw_bit)) | (rw_data << rw_bit); 

   // Write the replacement if write-enable is high
   always @(posedge clk) begin
      if(rw_we) begin
//	 $display("replacing mem[%d] = %b with %b", rw_line,
//		  memory[rw_line], rw_replace);

	 memory[rw_line] = rw_replace;
//	 $display("testreg is %b", testreg);
	 
      end // if (rw_we)
   end

   // TODO: short-circuit written data out of the read port.
   // Presently, data written propagates out after a positive clock edge,
   // and because mazekeeper holds outputs in registers, it actually
   // takes two clock cycles for data to come back.

endmodule // maze_dp_rowmem
